/*
    MorAnts (v1.0) - ANTS GAME bot 
    Copyright (C) 2011  Antonio M. Mora Garca

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/
   
*/           

#include "ants.h"

/* Constants for States */

#define ANTS_ATTACK 8   // global amount of ants considered to attack enemy hills
#define PROB_CLOSE 5    // indicates 1 out of PROB_CLOSE probability of moving close to another ant
#define TURNS_WITHOUT_GROW 20  // indicates how many turns should be the ants without change before change to explore mode

/* Local Variables */

/* positions to move in the turn */
struct move {
    int row;        // row to move on
    int col;        // column to move on
    int ant;        // ant index
};

struct move *moves;     // moves to perform
int movements_num;      // number of movements in the table


/* explorers */
struct explorer {
    int row;        // row to move on
    int col;        // column to move on
    char dir;       // current direction to follow
};

/* enemy and own hills */
struct basic_hill {
    int row;
    int col;
    int active;
    char player;
};

struct basic_hill *enemy_hills;
int num_enemy_hills = 0;

struct basic_hill *own_hills;
int num_own_hills = 0;


struct explorer *explorers;
int num_explorers = 0;

char last_dir;
int change_dir_attemps = 10;   /* number of attempts of changing direction */
int turn = 1;

int my_player_number;     // my hill number

int ants_amount = 0;      // number of ants counter
int not_grows = 0;

// -----------------------------------------------------------------------------

/* calculate the next position for the ant knowing the direction to move */
struct move calculate_next_position(struct game_info *Info, struct my_ant ant_moving, char direction) {
         
        struct move next_position;

        next_position.ant = ant_moving.id;
        
        if (direction == 'N') {
            if (ant_moving.row > 0)
               next_position.row = ant_moving.row-1;
            else
               next_position.row = Info->rows-1;
               
            next_position.col = ant_moving.col;
        }
        else if (direction == 'S') {
            if (ant_moving.row < (Info->rows-1))
               next_position.row = ant_moving.row+1;
            else
               next_position.row = 0;
                           
            next_position.col = ant_moving.col;
        }
        else if (direction == 'E') {
            next_position.row = ant_moving.row;
            
            if (ant_moving.col < (Info->cols-1))
               next_position.col = ant_moving.col+1;
            else
               next_position.col = 0;
        }
        else if (direction == 'W') {
            next_position.row = ant_moving.row;

            if (ant_moving.col > 0)
               next_position.col = ant_moving.col-1;
            else
               next_position.col = Info->cols-1;            
        }
        
        return next_position;
}

// -----------------------------------------------------------------------------

/* Check if a desired move is not previously included in the movement array.
   Moreover it should not be the current position of any ant.
*/
int possible_move (struct game_state *Game, struct game_info *Info, struct move *movs, struct move next_move){
      int i;
      
      /* check if it is in the next movement list */
      for (i=0;i<movements_num;i++)
          if ((movs[i].row == next_move.row) && (movs[i].col == next_move.col))
             return 0;
          
      /* check if it is in the current positions */    
      for (i=0;i<Game->my_count;i++)
          /* it's in the next movement list */
          if ((Game->my_ants[i].row == next_move.row) && (Game->my_ants[i].col == next_move.col))
             return 0;
      
      /* check the obstacles */
      if (Info->map[((next_move.row*Info->cols) + next_move.col)] == '%'){
         /* obstacle LOG */
         FILE *flog;
         flog = fopen("log.txt","a");                                           
         fprintf(flog,"\nOBST [%d,%d]",next_move.row,next_move.col);
         fclose(flog);
         return -1;
      }
          
      return 1;
}

// -----------------------------------------------------------------------------

/* calculate the best direction to move in order to get a position starting in another one */
char go_to(struct move orig_pos, struct move dest_pos, struct game_state *Game, struct game_info *Info) {
        
        int close_prob, i;
        int ant_dist, min_ant_dist = 65000;
        
        
        /* it tries firstly to approach other ants (randomly) - with a probability */
        close_prob = (rand()%PROB_CLOSE);
        
        /* moves close to the closest ant */
        if (close_prob<1) {
                          
             for (i=0; i<Game->my_count; i++) {
                 
                 if ((orig_pos.row == Game->my_ants[i].row) &&
                     (orig_pos.row == Game->my_ants[i].row))
                     continue;
                 
                 ant_dist = distance(orig_pos.row, orig_pos.col,
                                     Game->my_ants[i].row, Game->my_ants[i].col, Info);

                 /* changes destination if is closer than 5 cells */
                 if ((ant_dist<min_ant_dist) && (ant_dist>1) && (ant_dist<5)) {
                   min_ant_dist =  ant_dist;
                   dest_pos.row = Game->my_ants[i].row;
                   dest_pos.col = Game->my_ants[i].col;
                 }
             
             }
        
        }
        
        /* moves to destination */
        if (dest_pos.row < orig_pos.row)
            return 'N';
        else if (dest_pos.row > orig_pos.row) 
            return 'S';
        else if ((dest_pos.row == orig_pos.row) &&
                 (dest_pos.col > orig_pos.col)) 
            return 'E';
        else if ((dest_pos.row == orig_pos.row) &&
                 (dest_pos.col < orig_pos.col)) 
            return 'W';
        
}

// -----------------------------------------------------------------------------

/* random direction (except the one passed) */
char go_pseudo_random(char no_dir, int change_attemps) {
     
     int direc;
     
     do {
        direc = rand()%4;   
     
        if (direc == 0)
           return 'N';
        else if (direc == 1) 
           return 'S';
        else if (direc == 2) 
           return 'E';
        else if (direc == 3) 
           return 'W';
     } while ((direc == no_dir) && (change_attemps>0));
        
}

// -----------------------------------------------------------------------------

/* moving in a direction depending on its order */
char explore_grouped(struct game_state *Game, int ant_ind) {
     
     int direc, i;
     int group_size;
     
     /* if the ant is already in the explorers group, its direction is returned */
     for (i=0; i<num_explorers; i++) {
         if ((explorers[i].row == Game->my_ants[ant_ind].row) &&
             (explorers[i].col == Game->my_ants[ant_ind].col)) {
             
             /* direction LOG */
             FILE *flog;
             flog = fopen("log.txt","a");
             fprintf(flog,"\nturn %d - a(%d)[%d,%d]: explorer[%d] -> %c",turn,ant_ind,
                     explorers[i].row,explorers[i].col,i,explorers[i].dir);
             fclose(flog);
     
             return explorers[i].dir;
         }
     }
     
     /* number of ants in the group (each group moves in one direction) */
     group_size = (Game->my_count/4);
     
     /* direction (group) to move to depending on the ant index */
     direc = (int) floor((float)ant_ind/(float)group_size);
         
     /* direction LOG */
     FILE *flog;
     flog = fopen("log.txt","a");
     fprintf(flog,"\nturn %d - a(%d): group size = %d direc = %d",turn,ant_ind,group_size,direc);
     fclose(flog);          
          
     if (direc == 0)
         return 'N';
     else if (direc == 1) 
         return 'S';
     else if (direc == 2) 
         return 'E';
     else if (direc >= 3) 
         return 'W';
        
}

// -----------------------------------------------------------------------------

/* search for the ant in the explorers array and updates its coordinates */
void update_explorers(struct my_ant ant, char direc, struct move next_pos) {

     int i;

     /* search for the ant in the explorers group */
     for (i=0; i<num_explorers; i++) {
         if ((explorers[i].row == ant.row) &&
             (explorers[i].col == ant.col)) {
             explorers[i].row = next_pos.row;
             explorers[i].col = next_pos.col;

              /* LOG explorer found */
              FILE *flog;
              flog = fopen("log.txt","a");
              fprintf(flog,"\nturn %d - Explorer FOUND e(%d): [%d,%d]->[%d,%d]",turn,i,
                      ant.row,ant.col,next_pos.row,next_pos.col);
              fclose(flog);
              
             return;                              
         }

     }
     
     /* if it is not found, it saves the explorer */
     /* memory reservation for explorers */
     if (num_explorers>0) {
       explorers = realloc(explorers, ((num_explorers+1) * sizeof(struct explorer)));                   
     } else {                   
       free(explorers);
       explorers = malloc(sizeof(struct explorer));
     }
     
     /* save the explorer */
     explorers[num_explorers].row = next_pos.row;
     explorers[num_explorers].col = next_pos.col;
     explorers[num_explorers].dir = direc;

              /* LOG explorer saved */
              FILE *flog;
              flog = fopen("log.txt","a");
              fprintf(flog,"\nturn %d - Explorer SAVED e(%d): [%d,%d] - %c",turn,num_explorers,
                      explorers[num_explorers].row,explorers[num_explorers].col,explorers[num_explorers].dir);
              fclose(flog);

     num_explorers++;
        
}

// -----------------------------------------------------------------------------

/* returns 1 if there is any enemy hill */
int exist_enemy_hills (struct game_state *Game) {
 
     int i,j;
     int already_saved;
     
     for (i=0; i < Game->hill_count; i++) {

         /* save the enemy hills found */
         if (Game->hill[i].player != my_player_number) {
            
            already_saved = 0;
              
            /* check if the hill is already saved */
            for (j=0; j < num_enemy_hills; j++) {
                if ((enemy_hills[j].row == Game->hill[i].row) &&
                    (enemy_hills[j].col == Game->hill[i].col)) {
                    already_saved = 1;
                    break;
                }
            }
            if (already_saved) continue;  // try another hill
            
            /* memory reservation for enemy hills */  
            if (num_enemy_hills == 0) {
               free(enemy_hills);                 
               enemy_hills = malloc (sizeof(struct basic_hill));
            } else
               enemy_hills = realloc(enemy_hills, ((num_enemy_hills+1) * sizeof (struct basic_hill)));
       
            /* copy the data of the hill */     
            enemy_hills[num_enemy_hills].row = Game->hill[i].row;
            enemy_hills[num_enemy_hills].col = Game->hill[i].col;
            enemy_hills[num_enemy_hills].active = 1;
            enemy_hills[num_enemy_hills].player = Game->hill[i].player;
            
            num_enemy_hills++;

            /* enemy hills LOG */
            FILE *flog;
            flog = fopen("log.txt","a");
            fprintf(flog,"enemy hill [%d,%d]",enemy_hills[num_enemy_hills-1].row, enemy_hills[num_enemy_hills-1].col);
            fclose(flog);
       
            return 1;
         }
     }
     
     return num_enemy_hills;
    
}

// -----------------------------------------------------------------------------

void do_turn(struct game_state *Game, struct game_info *Info) {

    int i,j;
    int food_dist, min_food_dist;     // distance to the food
    int hill_dist, min_hill_dist;     // distance to the hills
    int enemy_dist, min_enemy_dist;   // distance to the enemies
    struct move food_pos, hill_pos, enemy_pos, ant_pos;
    char dir=-1;
    int can_move;
    struct move next_pos;  // next position
    FILE *flog;
    
    /* check the number of ants (should be increased) */
    if (Game->my_count == ants_amount) {
        not_grows++;                   
    } else {
        ants_amount = Game->my_count;
        not_grows = 0;
    }


    /* in the first turns I take not eof my player number */
    if ((turn<=2) && (Game->hill_count>0)) {
       my_player_number = Game->hill[0].player;
       
       if (num_own_hills == 0) {
          own_hills = malloc (Game->hill_count * sizeof(struct basic_hill));

          for (i=0; i<Game->hill_count; i++) {
              own_hills[i].row = Game->hill[0].row;
              own_hills[i].col = Game->hill[0].col;
              own_hills[i].active = 1;
              own_hills[i].player = Game->hill[0].player;
              num_own_hills++;
          }
          
       }
       
       /* Player ID and own hills LOG */
       flog = fopen("log.txt","a");
       fprintf(flog,"I AM PLAYER %d - Own Hills = %d",my_player_number,num_own_hills);
       fclose(flog); 

    }
        
    /* memory reservation for list of moves */
    moves = malloc(Game->my_count * sizeof(struct move));
    movements_num = 0;

    /* -----------------------  DECIDING ---------------------- */

    /* the ants are moved */
    for (i = 0; i < Game->my_count; ++i) {

        min_food_dist = 65000;
        min_hill_dist = 65000;
        min_enemy_dist = 65000;

        /***************************************************/
        /*                  HILL DEFENSE                   */
        /***************************************************/

        // To be implemented

        /***************************************************/
        /*                FOOD SEARCHING                   */
        /***************************************************/

        /* look for food */
        if ((Game->food_count) && (Game->my_count <= ANTS_ATTACK) && (not_grows<TURNS_WITHOUT_GROW)) {
        
             /* delete explorers */
             num_explorers = 0;
        
             /* current ant position */
             ant_pos.row = Game->my_ants[i].row;
             ant_pos.col = Game->my_ants[i].col;
 
             /* search for the closest food */
             for (j=0; j < Game->food_count; j++) {
                food_dist = distance(Game->my_ants[i].row, Game->my_ants[i].col,
                                   Game->food[j].row, Game->food[j].col, Info);
                if (food_dist<min_food_dist) {
                   min_food_dist =  food_dist;
                   food_pos.row = Game->food[j].row;
                   food_pos.col = Game->food[j].col;
                }
            } 
         
            dir = go_to(ant_pos,food_pos,Game,Info);
        
        /***************************************************/
        /*               HILLS ATTACKING                   */
        /***************************************************/
        
        /* if there are enemy hills, the ants attack them */  
        } else if (exist_enemy_hills(Game) && (not_grows<TURNS_WITHOUT_GROW)) {
        
             /* delete explorers */
             num_explorers = 0;
        
             /* current ant position */
             ant_pos.row = Game->my_ants[i].row;
             ant_pos.col = Game->my_ants[i].col;
 
             /* search for the closest hill */
             for (j=0; j < num_enemy_hills; j++) {
                 
                /* not consider non-active hills */
                if (enemy_hills[j].active == 0) 
                   continue;
                
                hill_dist = distance(Game->my_ants[i].row, Game->my_ants[i].col,
                                   enemy_hills[j].row, enemy_hills[j].col, Info);
                if (hill_dist<min_hill_dist) {
                   min_hill_dist =  hill_dist;
                   hill_pos.row = Game->hill[j].row;
                   hill_pos.col = Game->hill[j].col;
                }
            } 
         
            dir = go_to(ant_pos,hill_pos,Game,Info);        
        
         }
         
        /***************************************************/
        /*                ENEMY APPROACHING                */
        /***************************************************/
        else if ((Game->enemy_count>0) && (Game->my_count > ANTS_ATTACK) && (not_grows<TURNS_WITHOUT_GROW)) {
             
             /* delete explorers */
             num_explorers = 0;
             
             /* current ant position */
             ant_pos.row = Game->my_ants[i].row;
             ant_pos.col = Game->my_ants[i].col;
 
             /* search for the closest enemy */
             for (j=0; j < Game->enemy_count; j++) {
                                 
                enemy_dist = distance(Game->my_ants[i].row, Game->my_ants[i].col,
                                      Game->enemy_ants[j].row, Game->enemy_ants[j].col, Info);
                if (enemy_dist < min_enemy_dist) {
                   min_enemy_dist =  enemy_dist;
                   enemy_pos.row = Game->enemy_ants[j].row;
                   enemy_pos.col = Game->enemy_ants[j].col;
                   
                   /* check the orientation */
                   
                }
            } 
         
            dir = go_to(ant_pos,enemy_pos,Game,Info);        
        }

        /***************************************************/
        /*                   EXPLORE                       */
        /***************************************************/

        /* if no food, nor hills, the ant explores in a direction depending on its number */
        else {
             dir = explore_grouped(Game,i);
             
        }

        /* -----------------------  MOVING ---------------------- */

        /* calculate the next position for the ant */
        next_pos = calculate_next_position(Info,Game->my_ants[i],dir);
        
        /* check if the movement is possible (if there is no other ant moving 
        on the same position), otherwise the movement is not performed */
        
        can_move = possible_move(Game, Info, moves, next_pos);
        
        if (can_move==1) {
                
           // the ant do the movement
           if (dir != -1) {
                   
              /* movements LOG */
              FILE *flog;
              flog = fopen("log.txt","a");
              fprintf(flog,"\nturn %d - MOVE a(%d): [%d,%d]->[%d,%d]",turn,i,
                  Game->my_ants[i].row,Game->my_ants[i].col,next_pos.row,next_pos.col);
              fclose(flog);     
              
              /* if they are exploring (no known food), the exploring coordinates are changed */
              if (Game->food_count == 0) {
                 update_explorers(Game->my_ants[i],dir,next_pos);
              }
              
              /* the ant moves */     
              move(i, dir, Game, Info);
            
              /* store the new movement */    
              moves[movements_num].row=next_pos.row;
              moves[movements_num].col=next_pos.col;
              movements_num++;
              
           }

        /* NO MOVE: store the current position as the next position for the ant (avoiding to be collisioned) */
        } else if (can_move==0){
             moves[movements_num].row=Game->my_ants[i].row;
             moves[movements_num].col=Game->my_ants[i].col;
             movements_num++;
             
             /* NO movement LOG */
             FILE *flog;
             flog = fopen("log.txt","a");
             fprintf(flog,"\nturn %d - NOT MOVE a(%d): [%d,%d]",turn,i,
                     Game->my_ants[i].row,Game->my_ants[i].col);
             fclose(flog);
                  
        /* OBSTACLE: try another direction randomly */
        } else if (can_move == -1) {
            dir = go_pseudo_random(dir, 1);
                        
            /* calculate the next position for the ant */
            next_pos = calculate_next_position(Info,Game->my_ants[i],dir);
            
            if (possible_move(Game, Info, moves, next_pos)==1) {
              /* movements LOG */
              FILE *flog;
              flog = fopen("log.txt","a");
              fprintf(flog,"\nturn %d - AVOID OBSTACLE a(%d): [%d,%d]->[%d,%d]",turn,i,
                  Game->my_ants[i].row,Game->my_ants[i].col,next_pos.row,next_pos.col);
              fclose(flog);     
              
              /* if they are exploring (no known food), the exploring coordinates are changed */
              if (Game->food_count == 0) {
                 update_explorers(Game->my_ants[i],dir,next_pos);
              }
              
              /* the ant moves */     
              move(i, dir, Game, Info);
            
              /* store the new movement */    
              moves[movements_num].row=next_pos.row;
              moves[movements_num].col=next_pos.col;
              movements_num++;
            } else {
                 moves[movements_num].row=Game->my_ants[i].row;
                 moves[movements_num].col=Game->my_ants[i].col;
                 movements_num++;
             
                 /* NO movement LOG */
                 FILE *flog;
                 flog = fopen("log.txt","a");
                 fprintf(flog,"\nturn %d - NOT MOVE a(%d): [%d,%d]",turn,i,
                     Game->my_ants[i].row,Game->my_ants[i].col);
                     fclose(flog);
            
            }

        }
        
        /* if the ant has moved check if has destroyed a hill */
        if (can_move != 0) {
           
           /* if there was a hill, it is marked as non-active */
           for (j=0; j < num_enemy_hills; j++) {
             if ((enemy_hills[j].row == next_pos.row) && 
                 (enemy_hills[j].col == next_pos.col) &&
                 (enemy_hills[j].active == 1)) {
                 enemy_hills[j].active = 0;

                 /* enemy hills destroyed LOG */
                 FILE *flog;
                 flog = fopen("log.txt","a");
                 fprintf(flog,"\nenemy hill DESTROYED! [%d,%d]",enemy_hills[j].row, enemy_hills[j].col);
                 fclose(flog);
            
                 break;
             }
           }
           
        }
        
    } // for ants
    
    /* memory liberation */
    free(moves);

    turn++;
}
